package macros.rpc;

import haxe.macro.Context;
import haxe.macro.Expr;
using Lambda;

class RPC {
    macro static public function buildRPC(): Array<Field> {
        var fields = Context.getBuildFields();

        var className = Context.getLocalClass().get().name + "RPC";

        var c = macro class $className extends macros.rpc.RPCBase {
            public function new(id: kernel.net.Package.NetworkID) {
                super(id,$v{className});
            }
        }

        for (field in fields){
            if (field.meta == null) continue;
            if (field.meta.exists((i) -> i.name == "rpc") == false) continue;

            switch (field.kind){
                case FFun(f):
                    c.fields.push({
                       name: field.name,
                       pos: field.pos,
                       kind: FFun({
                           args: f.args,
                           expr: macro {
                               return cast this._performRequest($v{field.name},[]);
                           },
                           ret: TPath({name: "Promise", params: [TPType(f.ret)], pack: ["tink","core"]}),
                       }),
                       access: [APublic],
                       doc: null,
                       meta: [],
                   });
                default:
                    Context.error("Only functions can be used for rpc", field.pos);
            }
        }

        haxe.macro.Context.defineType(c);

        return fields;
    }

    macro static public function generateRPCPackageHandle() {
        var proto = Context.getLocalClass().get().name + "RPC";
        var exprs: Array<Expr> = [];

        var fields = Context.getLocalClass().get().fields.get();

        for (field in fields){
            if (field.meta == null) continue;
            if (!field.meta.has("rpc")) continue;

            switch (field.kind){
                case FMethod(k):
                    var funName = field.name;
                    exprs.push(macro {
                        if (pack.data.func == $v{funName}){
                            pack.respond(this.$funName());
                        }
                    });

                default:
                    Context.error("Only functions can be used for rpc", field.pos);
            }
        }

        return macro {
            kernel.net.Net.instance.registerProto($v{proto},(pack)->{
                $a{exprs}
            });
        };
    }
}
